// -*- C++ -*-

/* 
 * GChemPaint
 * atom.h 
 *
 * Copyright (C) 2001-2003
 *
 * Developed by Jean Brfort <jean.brefort@ac-dijon.fr>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifndef GCHEMPAINT_ATOM_H
#define GCHEMPAINT_ATOM_H

#include "config.h"
#include <map>
#include <glib.h>
#ifdef HAVE_OPENBABEL
#	include <mol.h>
using namespace OpenBabel;
#endif
#include <chemistry/atom.h>

using namespace std;
using namespace gcu;

class gcpBond;
class gcpCycle;

class gcpAtom: public Atom
{
public:
	gcpAtom();
	gcpAtom(int Z, double x, double y, double z);
#ifdef HAVE_OPENBABEL
	gcpAtom(OBAtom* atom, double zoom = 1.0);
#endif
	virtual ~gcpAtom();

public :
	double Distance(gcpAtom* pAtom);
	void zoom(double ZoomFactor);
	virtual void SetZ(int Z);
	void AddBond(gcpBond* pBond);
	void RemoveBond(gcpBond* pBond);
	void Update();
	virtual void Add(GtkWidget* w);
	virtual void Update(GtkWidget* w);
	int GetTotalBondsNumber(); //take bond order into account
	int GetAttachedHydrogens() {return m_nH;}
	bool GetBestSide();
	bool IsInCycle(gcpCycle* pCycle);
	int GetChargePosition(double& x, double& y);
	virtual bool LoadNode(xmlNodePtr);

private:
	int m_nH;
	int m_Valence, m_nve; //valence, number of valence electrons
	double m_width, m_height; //size of the atomic symbol in the canvas
	bool m_HPos; //false = left, true = right
	bool m_ChargeAuto;
};

#endif // GCHEMPAINT_ATOM_H
