// -*- C++ -*-

/* 
 * GChemPaint
 * text.cc 
 *
 * Copyright (C) 2002-2003
 *
 * Developed by Jean Bréfort <jean.brefort@ac-dijon.fr>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "config.h"
#include "text.h"
#include "widgetdata.h"
#include "view.h"
#include "document.h"
#include "settings.h"
#include "globals.h"

extern GtkTextTagTable* TextTagTable;

/*
Derivation of a new widget from GnomeCanvasRichText with an event for setting canvas size on realize
*/

typedef struct _GnomeCanvasRichTextGCP           GnomeCanvasRichTextGCP;
typedef struct _GnomeCanvasRichTextGCPClass      GnomeCanvasRichTextGCPClass;

#define GNOME_TYPE_CANVAS_RICH_TEXT_GCP            (gnome_canvas_rich_text_gcp_get_type ())
#define GNOME_CANVAS_RICH_TEXT_GCP(obj)            (GTK_CHECK_CAST ((obj), GNOME_TYPE_CANVAS_RICH_TEXT_GCP, GnomeCanvasRichTextGCP))
#define GNOME_CANVAS_CLASS_RICH_TEXT_GCP(klass)    (GTK_CHECK_CLASS_CAST ((klass), GNOME_TYPE_CANVAS_RICH_TEXT_GCP, GnomeCanvasRichTextGCPClass))
#define GNOME_IS_CANVAS_RICH_TEXT_GCP(obj)         (GTK_CHECK_TYPE ((obj), GNOME_TYPE_CANVAS_RICH_TEXT_GCP))
#define GNOME_IS_CANVAS_RICH_TEXT_GCP_CLASS(klass) (GTK_CHECK_CLASS_TYPE ((klass), GNOME_TYPE_CANVAS_RICH_TEXT_GCP))
#define GNOME_CANVAS_RICH_TEXT_GCP_GET_CLASS(obj)  (GTK_CHECK_GET_CLASS ((obj), GNOME_TYPE_CANVAS_RICH_TEXT_GCP, GnomeCanvasRichTextGCPClass))


enum {
  UPDATE_BOUNDS,
  LAST_SIGNAL
};

struct _GnomeCanvasRichTextGCP {
	GnomeCanvasRichText canvas;
};

struct _GnomeCanvasRichTextGCPClass {
	GnomeCanvasRichTextClass parent_class;
};


GType gnome_canvas_rich_text_gcp_get_type (void) G_GNUC_CONST;
static void gnome_canvas_rich_text_gcp_class_init(GnomeCanvasRichTextGCPClass *Class);
static void gnome_canvas_rich_text_gcp_init(GnomeCanvasRichTextGCP *canvas);
static GtkWidget *gnome_canvas_rich_text_gcp_new (void);

static void gnome_canvas_rich_text_gcp_realize(GnomeCanvasItem *item);
static GnomeCanvasItemClass *parent_class;

GType
gnome_canvas_rich_text_gcp_get_type (void)
{
	static GType canvas_rich_text_gcp_type;

	if (!canvas_rich_text_gcp_type) {
		static const GTypeInfo object_info = {
			sizeof(GnomeCanvasRichTextGCPClass),
			(GBaseInitFunc)NULL,
			(GBaseFinalizeFunc)NULL,
			(GClassInitFunc)gnome_canvas_rich_text_gcp_class_init,
			(GClassFinalizeFunc)NULL,
			NULL,			/* class_data */
			sizeof(GnomeCanvasRichTextGCP),
			0,			/* n_preallocs */
			(GInstanceInitFunc)gnome_canvas_rich_text_gcp_init,
			NULL			/* value_table */
		};

		canvas_rich_text_gcp_type = g_type_register_static(GNOME_TYPE_CANVAS_RICH_TEXT, "GnomeCanvasRichTextGCP",
						      &object_info, (GTypeFlags)0);
	}

	return canvas_rich_text_gcp_type;
}

void gnome_canvas_rich_text_gcp_class_init(GnomeCanvasRichTextGCPClass *Class)
{
	GnomeCanvasItemClass *item_class = GNOME_CANVAS_ITEM_CLASS(Class);
	parent_class = (GnomeCanvasItemClass*)g_type_class_peek_parent (Class);
	item_class->realize = (void (*)(GnomeCanvasItem*))gnome_canvas_rich_text_gcp_realize;
}

void gnome_canvas_rich_text_gcp_init(GnomeCanvasRichTextGCP *canvas)
{
}

static void gnome_canvas_rich_text_gcp_realize(GnomeCanvasItem *item)
{
	GnomeCanvasRichText *text = GNOME_CANVAS_RICH_TEXT(item);
	(* GNOME_CANVAS_ITEM_CLASS(parent_class)->realize)(item);
	gcpText* pText = (gcpText*)g_object_get_data(G_OBJECT(item), "object");
	g_object_set_data(G_OBJECT(item), "realized", (void*)1);
	if (pText) pText->OnChanged(gnome_canvas_rich_text_get_buffer(text));
}

static bool on_changed(GtkTextBuffer *textbuffer, gcpText *text)
{
	return text->OnChanged(textbuffer);
}

static bool on_mark_set(GtkTextBuffer *textbuffer, GtkTextIter *iter, GtkTextMark *mark, gcpText *text)
{
	return text->OnMarkSet(textbuffer, iter, mark);
}

static bool on_insert_text(GtkTextBuffer *textbuffer, GtkTextIter *iter, gchar* newtext, gint length, gcpText *text)
{
	return text->OnInsertText(textbuffer, iter, newtext, length);
}

static bool on_end_user_action(GtkTextBuffer *textbuffer, gcpText *text)
{
	return text->OnEndUserAction(textbuffer);
}

gcpText::gcpText(): Object(TextType)
{
	m_x = 0;
	m_y = 0;
	m_length = 5;
	m_height = 15;
	m_buf = gtk_text_buffer_new(TextTagTable);
	g_signal_connect(G_OBJECT(m_buf), "changed", G_CALLBACK(on_changed), this);
	g_signal_connect(G_OBJECT(m_buf), "mark-set", G_CALLBACK(on_mark_set), this);
	g_signal_connect(G_OBJECT(m_buf), "insert-text", G_CALLBACK(on_insert_text), this);
	g_signal_connect(G_OBJECT(m_buf), "end-user-action", G_CALLBACK(on_end_user_action), this);
	m_bLoading = false;
}

gcpText::gcpText(double x, double y): Object(TextType)
{
	m_x = x;
	m_y = y;
	m_length = 5;
	m_height = 15;
	m_buf = gtk_text_buffer_new(TextTagTable);
	g_signal_connect(G_OBJECT(m_buf), "changed", G_CALLBACK(on_changed), this);
	g_signal_connect(G_OBJECT(m_buf), "mark-set", G_CALLBACK(on_mark_set), this);
	g_signal_connect(G_OBJECT(m_buf), "insert-text", G_CALLBACK(on_insert_text), this);
	g_signal_connect(G_OBJECT(m_buf), "end-user-action", G_CALLBACK(on_end_user_action), this);
	m_bLoading = false;
}

gcpText::~gcpText()
{
	if (m_buf) g_object_unref((GObject*)m_buf);
}

void gcpText::GetCoords(double *x, double *y)
{
	*x = m_x;
	*y = m_y;
}

void gcpText::SetCoords(double x, double y)
{
	m_x = x;
	m_y = y;
}

xmlNodePtr gcpText::Save(xmlDocPtr xml)
{
	GtkTextIter start, end;
	gtk_text_buffer_get_start_iter(m_buf, &start);
	gtk_text_buffer_get_end_iter(m_buf, &end);
	xmlNodePtr node = xmlNewDocNode(xml, NULL, (xmlChar*)"text", NULL);
	if (!node) return NULL;
	GSList* TagList = gtk_text_iter_get_tags(&start), *OtherTagList = NULL;
	if (! SaveNode(xml, node, &start, &end, TagList, OtherTagList))
	{
		xmlFree(node);
		return NULL;
	}
	g_slist_free(TagList);
	if (m_Id) xmlNewProp(node, (xmlChar*)"id", (xmlChar*)m_Id);
	char buf[16];
	g_snprintf(buf, sizeof(buf), "%g", m_x);
	xmlNewProp(node, (xmlChar*)"x", (xmlChar*)buf);
	g_snprintf(buf, sizeof(buf), "%g", m_y);
	xmlNewProp(node, (xmlChar*)"y", (xmlChar*)buf);
	return node;
}

bool gcpText::SaveNode(xmlDocPtr xml, xmlNodePtr node, GtkTextIter *iter, GtkTextIter *end, GSList*& TagList, GSList*& OtherTagList)
{
	GtkTextIter start = *iter, next;
	GSList* TagList0, *TagList1;
	if (g_slist_length(TagList) == 0)
	{
		gtk_text_iter_forward_to_tag_toggle(iter, NULL);
		char *text = gtk_text_buffer_get_text(m_buf, &start, iter, true);
		xmlNodeAddContent(node, (xmlChar*)text);
	}
	else
	{
		//find the longest included tag
		TagList0 = TagList;
		GtkTextTag *tag;
		while (TagList0)
		{
			next = start;
			gtk_text_iter_forward_to_tag_toggle(&next, GTK_TEXT_TAG(TagList0->data));
			if (gtk_text_iter_compare(iter, &next) < 0)
			{
				*iter = next;
				tag = GTK_TEXT_TAG(TagList0->data);
			}
			if (gtk_text_iter_compare(end, &next) <= 0)
			{
				*iter = *end;
				break;
			}
			TagList0 = TagList0->next;
		}
		GValue val;
		memset(&val, 0, sizeof(GValue));
		gchar* tagname;
		g_value_init(&val, G_TYPE_STRING);
		g_object_get_property((GObject*) tag, "name", &val);
		const gchar* name = g_value_get_string(&val);
		if (!strcmp(name, "bold")) tagname = "b";
		else if  (!strcmp(name, "italic")) tagname = "i";
		else if  (!strcmp(name, "underline")) tagname = "u";
		else if  (!strcmp(name, "strikethrough")) tagname = "s";
		else if  (!strcmp(name, "subscript")) tagname = "sub";
		else if  (!strcmp(name, "superscript")) tagname = "sup";
		OtherTagList = g_slist_append(OtherTagList, tag);
		TagList = g_slist_remove(TagList, tag);
		xmlNodePtr child = xmlNewDocNode(xml, NULL, (xmlChar*)tagname, NULL);
		if (!child) return false;
		SaveNode(xml, child, &start, iter, TagList, OtherTagList);
		xmlAddChild(node, child);
		OtherTagList = g_slist_remove(OtherTagList, tag);
		TagList = g_slist_append(TagList, tag);
	}
	if (!gtk_text_iter_equal(iter, end))
	{
		TagList1 = TagList0 = gtk_text_iter_get_tags(iter);
		if (OtherTagList) while (TagList1)
		{
			if (g_slist_find(OtherTagList, TagList1->data)) TagList0 = g_slist_remove(TagList0, TagList1->data);
			TagList1 = TagList1->next;
		}
		SaveNode(xml, node, iter, end, TagList0, OtherTagList);
		g_slist_free(TagList0);
	}
	return true;
}

bool gcpText::Load(xmlNodePtr node)
{
	char* tmp, *endptr;
	xmlNodePtr child;
	tmp = (char*)xmlGetProp(node, (xmlChar*)"id");
	if (tmp) SetId(tmp);
	tmp = (char*)xmlGetProp(node, (xmlChar*)"x");
	if (!tmp) return false;
	m_x = strtod(tmp, &endptr);
	if (*endptr) return false;
	tmp = (char*)xmlGetProp(node, (xmlChar*)"y");
	if (!tmp) return false;
	m_y = strtod(tmp, &endptr);
	if (*endptr) return false;
	m_bLoading = true;
	child = node->children;
	GtkTextIter iter;
	gtk_text_buffer_get_iter_at_offset (m_buf, &iter, 0);
	while (child)
	{
		LoadNode(child, &iter);
		child = child->next;
	}
	m_bLoading = false;
}

bool gcpText::LoadNode(xmlNodePtr node, GtkTextIter* CurIter)
{
	char* tmp;
	if (!strcmp((const char*)node->name, "text"))
	{
		tmp = (char*) xmlNodeGetContent(node);
		gtk_text_buffer_insert(m_buf, CurIter, tmp, -1);
		return true;
	}
	else if (!strcmp((const char*)node->name, "b")) return LoadTaggedNode(node, CurIter, "bold");
	else if (!strcmp((const char*)node->name, "i")) return LoadTaggedNode(node, CurIter, "italic");
	else if (!strcmp((const char*)node->name, "u")) return LoadTaggedNode(node, CurIter, "underline");
	else if (!strcmp((const char*)node->name, "s")) return LoadTaggedNode(node, CurIter, "strikethrough");
	else if (!strcmp((const char*)node->name, "sub")) return LoadTaggedNode(node, CurIter, "subscript");
	else if (!strcmp((const char*)node->name, "sup")) return LoadTaggedNode(node, CurIter, "superscript");
}

bool gcpText::LoadTaggedNode(xmlNodePtr node, GtkTextIter* CurIter, const char* TagName)
{
	GtkTextIter start, end;
	gint offset = gtk_text_iter_get_offset(CurIter);
	xmlNodePtr child = node->children;
	while (child)
	{
		LoadNode(child, CurIter);
		child = child->next;
	}
	gtk_text_buffer_get_iter_at_offset(m_buf, &start, offset);
	gtk_text_buffer_get_end_iter(m_buf, &end);
	gtk_text_buffer_apply_tag(m_buf, gtk_text_tag_table_lookup(TextTagTable, TagName), &start, &end);
}

void gcpText::Add(GtkWidget* w)
{
	gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(w), "data");
	GnomeCanvasGroup* group = GNOME_CANVAS_GROUP(gnome_canvas_item_new(pData->Group, gnome_canvas_group_get_type(), NULL));
	GnomeCanvasItem* item = gnome_canvas_item_new(
						group,
						gnome_canvas_rect_get_type(),
						"x1", m_x * pData->ZoomFactor - pData->Padding,
						"y1", m_y * pData->ZoomFactor - pData->Padding,
						"x2", m_x * pData->ZoomFactor + m_length + pData->Padding,
						"y2", m_y * pData->ZoomFactor + m_height + pData->Padding,
						"fill_color", "white",
						"outline_color", "white",
						NULL);
	g_object_set_data(G_OBJECT(group), "rect", item);
	g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
	g_object_set_data(G_OBJECT(item), "object", this);
	item = gnome_canvas_item_new(
						group,
						gnome_canvas_rich_text_gcp_get_type(),
						"text", "",
						"x", m_x * pData->ZoomFactor,
						"y", m_y * pData->ZoomFactor,
						"width", m_length,
						"height", m_height,
						"grow_height", false,
						"editable", false,
						"cursor_visible", false,
						NULL);
	g_object_set_data(G_OBJECT(group), "text", item);
	gnome_canvas_rich_text_set_buffer(GNOME_CANVAS_RICH_TEXT(item), m_buf);
	g_object_set_data(G_OBJECT(item), "object", this);
	g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), w);
	pData->Items[this] = group;
	if (g_object_get_data(G_OBJECT(item), "realized")) OnChanged(m_buf);
}

void gcpText::Move(double x, double y, double z)
{
	m_x += x;
	m_y += y;
}

bool gcpText::OnChanged(GtkTextBuffer *textbuffer)
{
	if (m_bLoading) return false;
	if (textbuffer != m_buf) return false;
	gcpDocument* pDoc = (gcpDocument*)GetDocument();
	if (!pDoc) return false;
	gcpView* pView = pDoc->GetView();
	GtkWidget* pWidget = pView->GetWidget();
	gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(pWidget), "data");
	GnomeCanvasRichText* text = (GnomeCanvasRichText*)g_object_get_data(G_OBJECT(pData->Items[this]), "text");
	GdkRectangle rect;
	GtkTextIter iter;
	gtk_text_buffer_get_end_iter(m_buf,&iter);
	gnome_canvas_rich_text_get_iter_location(text, &iter, &rect);
	m_length = rect.x + 10;
	if (m_height <= rect.height) m_height = rect.height + 1;
	while (rect.y)
	{
		int y = rect.y;
		while (y == rect.y)
		{
			gtk_text_iter_backward_char(&iter);
			gnome_canvas_rich_text_get_iter_location(text, &iter, &rect);
		}
		m_length += rect.x + 3;//FIXME: add a space width instead of 3
	}
	pView->Update(this);
	return true;
}

bool gcpText::OnMarkSet(GtkTextBuffer *textbuffer, GtkTextIter *iter, GtkTextMark *mark)
{
	if (textbuffer != m_buf) return false;
	GtkTextIter start, end, next;
	gtk_text_buffer_get_selection_bounds(m_buf, &start, &end);
	next = start;
	GtkTextTag *tag = gtk_text_tag_table_lookup(TextTagTable, "bold");
	bool bTag = gtk_text_iter_has_tag(&start, tag);
	gtk_text_iter_forward_to_tag_toggle(&next, tag);
	gtk_toggle_button_set_active((GtkToggleButton*)BoldBtn, bTag);
	gtk_toggle_button_set_inconsistent((GtkToggleButton*)BoldBtn, gtk_text_iter_compare(&next, &end) < 0);
	next = start;
	tag = gtk_text_tag_table_lookup(TextTagTable, "italic");
	bTag = gtk_text_iter_has_tag(&start, tag);
	gtk_text_iter_forward_to_tag_toggle(&next, tag);
	gtk_toggle_button_set_active((GtkToggleButton*)ItalicBtn, bTag);
	gtk_toggle_button_set_inconsistent((GtkToggleButton*)ItalicBtn, gtk_text_iter_compare(&next, &end) < 0);
	next = start;
	tag = gtk_text_tag_table_lookup(TextTagTable, "underline");
	bTag = gtk_text_iter_has_tag(&start, tag);
	gtk_text_iter_forward_to_tag_toggle(&next, tag);
	gtk_toggle_button_set_active((GtkToggleButton*)UnderlineBtn, bTag);
	gtk_toggle_button_set_inconsistent((GtkToggleButton*)UnderlineBtn, gtk_text_iter_compare(&next, &end) < 0);
	next = start;
	tag = gtk_text_tag_table_lookup(TextTagTable, "strikethrough");
	bTag = gtk_text_iter_has_tag(&start, tag);
	gtk_text_iter_forward_to_tag_toggle(&next, tag);
	gtk_toggle_button_set_active((GtkToggleButton*)StrikethroughBtn, bTag);
	gtk_toggle_button_set_inconsistent((GtkToggleButton*)StrikethroughBtn, gtk_text_iter_compare(&next, &end) < 0);
	next = start;
	tag = gtk_text_tag_table_lookup(TextTagTable, "subscript");
	bTag = gtk_text_iter_has_tag(&start, tag);
	gtk_text_iter_forward_to_tag_toggle(&next, tag);
	gtk_toggle_button_set_active((GtkToggleButton*)SubscriptBtn, bTag);
	gtk_toggle_button_set_inconsistent((GtkToggleButton*)SubscriptBtn, gtk_text_iter_compare(&next, &end) < 0);
	next = start;
	tag = gtk_text_tag_table_lookup(TextTagTable, "superscript");
	bTag = gtk_text_iter_has_tag(&start, tag);
	gtk_text_iter_forward_to_tag_toggle(&next, tag);
	gtk_toggle_button_set_active((GtkToggleButton*)SuperscriptBtn, bTag);
	gtk_toggle_button_set_inconsistent((GtkToggleButton*)SuperscriptBtn, gtk_text_iter_compare(&next, &end) < 0);
}

bool gcpText::OnInsertText(GtkTextBuffer *textbuffer, GtkTextIter *iter, gchar* newtext, gint length)
{
	m_InsertOffset = gtk_text_iter_get_offset(iter);
	return true;
}

bool gcpText::OnEndUserAction(GtkTextBuffer *textbuffer)
{
	GtkTextIter start, end;
	gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end);
	if (gtk_text_iter_get_offset(&end) == m_InsertOffset + 1)
	{
		gtk_text_iter_backward_char(&start);
		m_InsertOffset = -2;
		if (gtk_toggle_button_get_active((GtkToggleButton*)BoldBtn)) gtk_text_buffer_apply_tag_by_name(textbuffer, "bold", &start, &end);
		if (gtk_toggle_button_get_active((GtkToggleButton*)ItalicBtn)) gtk_text_buffer_apply_tag_by_name(textbuffer, "italic", &start, &end);
		if (gtk_toggle_button_get_active((GtkToggleButton*)UnderlineBtn)) gtk_text_buffer_apply_tag_by_name(textbuffer, "underline", &start, &end);
		if (gtk_toggle_button_get_active((GtkToggleButton*)StrikethroughBtn)) gtk_text_buffer_apply_tag_by_name(textbuffer, "strikethrough", &start, &end);
		if (gtk_toggle_button_get_active((GtkToggleButton*)SubscriptBtn)) gtk_text_buffer_apply_tag_by_name(textbuffer, "subscript", &start, &end);
		if (gtk_toggle_button_get_active((GtkToggleButton*)SuperscriptBtn)) gtk_text_buffer_apply_tag_by_name(textbuffer, "superscript", &start, &end);
	}
	return true;
}

void gcpText::Update(GtkWidget* w)
{
	gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(w), "data");
	GnomeCanvasGroup* group = pData->Items[this];
	g_object_set(G_OBJECT(g_object_get_data(G_OBJECT(group), "text")),
						"x", m_x * pData->ZoomFactor,
						"y", m_y * pData->ZoomFactor,
						"width", m_length,
						"height", m_height,
						NULL);
	g_object_set(G_OBJECT(g_object_get_data(G_OBJECT(group), "rect")),
						"x1", m_x * pData->ZoomFactor - pData->Padding,
						"y1", m_y * pData->ZoomFactor - pData->Padding,
						"x2", m_x * pData->ZoomFactor + m_length + pData->Padding,
						"y2", m_y * pData->ZoomFactor + m_height + pData->Padding,
						NULL);
}

void gcpText::SetSelected(GtkWidget* w, int state)
{
	gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(w), "data");
	GnomeCanvasGroup* group = pData->Items[this];
	gchar* color;
	switch (state)
	{	
		case SelStateUnselected: color = "white"; break;
		case SelStateSelected: color = SelectColor; break;
		case SelStateUpdating: color = AddColor; break;
		case SelStateErasing: color = DeleteColor; break;
	}
	g_object_set(G_OBJECT(g_object_get_data(G_OBJECT(group), "rect")), "outline_color", color, NULL);
}
