// -*- C++ -*-

/* 
 * GChemPaint
 * view.cc
 *
 * Copyright (C) 2001-2003
 *
 * Developed by Jean Brfort <jean.brefort@ac-dijon.fr>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "config.h"
#include "widgetdata.h"
#include "view.h"
#include "settings.h"
#include "document.h"
#include "tool.h"
#include "globals.h"
#include "text.h"
#include <math.h>
#include <pango/pango-context.h>

/*
Derivation of a new widget from gnome_canvas with an event for updating canvas size
*/

typedef struct _GnomeCanvasGCP           GnomeCanvasGCP;
typedef struct _GnomeCanvasGCPClass      GnomeCanvasGCPClass;

#define GNOME_TYPE_CANVAS_GCP            (gnome_canvas_gcp_get_type ())
#define GNOME_CANVAS_GCP(obj)            (GTK_CHECK_CAST ((obj), GNOME_TYPE_CANVAS_GCP, GnomeCanvasGCP))
#define GNOME_CANVAS_CLASS_GCP(klass)    (GTK_CHECK_CLASS_CAST ((klass), GNOME_TYPE_CANVAS_GCP, GnomeCanvasGCPClass))
#define GNOME_IS_CANVAS_GCP(obj)         (GTK_CHECK_TYPE ((obj), GNOME_TYPE_CANVAS_GCP))
#define GNOME_IS_CANVAS_GCP_CLASS(klass) (GTK_CHECK_CLASS_TYPE ((klass), GNOME_TYPE_CANVAS_GCP))
#define GNOME_CANVAS_GCP_GET_CLASS(obj)  (GTK_CHECK_GET_CLASS ((obj), GNOME_TYPE_CANVAS_GCP, GnomeCanvasGCPClass))


enum {
  UPDATE_BOUNDS,
  LAST_SIGNAL
};

struct _GnomeCanvasGCP {
	GnomeCanvas canvas;
};

struct _GnomeCanvasGCPClass {
	GnomeCanvasClass parent_class;

	void (* update_bounds) (GnomeCanvasGCP *canvas);
};


GType gnome_canvas_gcp_get_type (void) G_GNUC_CONST;
static void gnome_canvas_gcp_class_init(GnomeCanvasGCPClass *Class);
static void gnome_canvas_gcp_init(GnomeCanvasGCP *canvas);
static void gnome_canvas_gcp_update_bounds(GnomeCanvasGCP *canvas);
static GtkWidget *gnome_canvas_gcp_new (void);

static guint gnome_canvas_gcp_signals[LAST_SIGNAL] = { 0 };

GType
gnome_canvas_gcp_get_type (void)
{
	static GType canvas_gcp_type;

	if (!canvas_gcp_type) {
		static const GTypeInfo object_info = {
			sizeof(GnomeCanvasGCPClass),
			(GBaseInitFunc)NULL,
			(GBaseFinalizeFunc)NULL,
			(GClassInitFunc)gnome_canvas_gcp_class_init,
			(GClassFinalizeFunc)NULL,
			NULL,			/* class_data */
			sizeof(GnomeCanvasGCP),
			0,			/* n_preallocs */
			(GInstanceInitFunc)gnome_canvas_gcp_init,
			NULL			/* value_table */
		};

		canvas_gcp_type = g_type_register_static(GNOME_TYPE_CANVAS, "GnomeCanvasGCP",
						      &object_info, (GTypeFlags)0);
	}

	return canvas_gcp_type;
}

GtkWidget *gnome_canvas_gcp_new(void)
{
	return GTK_WIDGET(g_object_new(GNOME_TYPE_CANVAS_GCP, /*"aa", TRUE,*/ NULL));
}

void gnome_canvas_gcp_class_init(GnomeCanvasGCPClass *Class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (Class);
	gnome_canvas_gcp_signals[UPDATE_BOUNDS] =
	g_signal_new ("update_bounds",
				  G_TYPE_FROM_CLASS(gobject_class),
				  G_SIGNAL_RUN_LAST,
				  G_STRUCT_OFFSET(GnomeCanvasGCPClass, update_bounds),
				  NULL, NULL,
				  g_cclosure_marshal_VOID__VOID,
				  G_TYPE_NONE, 0
				  );
	Class->update_bounds = gnome_canvas_gcp_update_bounds;
}

void gnome_canvas_gcp_init(GnomeCanvasGCP *canvas)
{
	;
}

void gnome_canvas_gcp_update_bounds(GnomeCanvasGCP *canvas)
{
	while (gtk_events_pending()) gtk_main_iteration();
	gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(canvas), "data");
	double x1, y1, x2, y2;
	gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(pData->Group), &x1, &y1, &x2, &y2);
	gcpView *pView = (gcpView*)g_object_get_data(G_OBJECT(canvas), "view");
	pView->UpdateSize(x1, y1, x2, y2);
}

bool on_event(GnomeCanvasItem *item, GdkEvent *event, GtkWidget* widget)
{
	gcpView* pView = (gcpView*) g_object_get_data(G_OBJECT(widget), "view");
	return pView->OnEvent(item, event, widget);
}

static bool on_destroy(GtkWidget *widget, gcpView * pView)
{
	pView->OnDestroy(widget);
	return true;
}

static bool on_size(GtkWidget *widget, GtkAllocation *alloc, gcpView * pView)
{
	pView->OnSize(alloc->width, alloc->height);
	return true;
}

using namespace std;

gcpView::gcpView(gcpDocument *pDoc)
{
	m_pDoc = pDoc;
	m_PangoFontDesc = pango_font_description_new();
	pango_font_description_set_family(m_PangoFontDesc, DefaultFontFamily);
	pango_font_description_set_style(m_PangoFontDesc, DefaultFontStyle);
	pango_font_description_set_weight(m_PangoFontDesc, DefaultFontWeight);
	pango_font_description_set_variant(m_PangoFontDesc, DefaultFontVariant);
	pango_font_description_set_stretch(m_PangoFontDesc, DefaultFontStretch);
	pango_font_description_set_size(m_PangoFontDesc, DefaultFontSize);
	m_sFontName = pango_font_description_to_string(m_PangoFontDesc);
	m_PangoSmallFontDesc = pango_font_description_new();
	pango_font_description_set_family(m_PangoSmallFontDesc, DefaultFontFamily);
	pango_font_description_set_style(m_PangoSmallFontDesc, DefaultFontStyle);
	pango_font_description_set_weight(m_PangoSmallFontDesc, DefaultFontWeight);
	pango_font_description_set_variant(m_PangoSmallFontDesc, DefaultFontVariant);
	pango_font_description_set_stretch(m_PangoSmallFontDesc, DefaultFontStretch);
	pango_font_description_set_size(m_PangoSmallFontDesc, DefaultFontSize * 2 / 3);
	m_sSmallFontName = pango_font_description_to_string(m_PangoSmallFontDesc);
	m_PangoTinyFontDesc = pango_font_description_new();
	pango_font_description_set_family(m_PangoTinyFontDesc, DefaultFontFamily);
	pango_font_description_set_style(m_PangoTinyFontDesc, DefaultFontStyle);
	pango_font_description_set_weight(m_PangoTinyFontDesc, DefaultFontWeight);
	pango_font_description_set_variant(m_PangoTinyFontDesc, DefaultFontVariant);
	pango_font_description_set_stretch(m_PangoTinyFontDesc, DefaultFontStretch);
	pango_font_description_set_size(m_PangoTinyFontDesc, DefaultFontSize / 2);
	m_sTinyFontName = pango_font_description_to_string(m_PangoTinyFontDesc);
	m_DefaultPadding = DefaultPadding;
	m_DefaultStereoBondWidth = DefaultStereoBondWidth;
	m_DefaultHashWidth = DefaultHashWidth;
	m_DefaultHashDist = DefaultHashDist;
	m_DefaultBondWidth = DefaultBondWidth;
	m_DefaultBondLength = DefaultBondLength;
	m_DefaultBondDist = DefaultBondDist;
	m_DefaultArrowLength = DefaultArrowLength;
	m_DefaultArrowHeadA = DefaultArrowHeadA;
	m_DefaultArrowHeadB = DefaultArrowHeadB;
	m_DefaultArrowHeadC = DefaultArrowHeadC;
	m_DefaultArrowDist = DefaultArrowDist;
	m_width = 400;
	m_height = 300;
	if (!IsEmbedded()) CreateNewWidget();
}

gcpView::~gcpView()
{
	if (!bCloseAll &&!IsEmbedded())
	{
		void* Item = g_object_get_data(G_OBJECT(m_pWidget), "menu");
		if (Item) gtk_widget_destroy(GTK_WIDGET(Item));
	}
	g_object_unref(G_OBJECT(m_PangoContext));
	if (m_sFontName) g_free(m_sFontName);
	if (m_sSmallFontName) g_free(m_sSmallFontName);
}

static bool bDragging = false;

bool gcpView::OnEvent(GnomeCanvasItem *item, GdkEvent *event, GtkWidget* widget)
{
	if (!pActiveTool) return true;
	if (item == (GnomeCanvasItem*)m_ActiveRichText)
	{
		GnomeCanvasItemClass* klass = GNOME_CANVAS_ITEM_CLASS(((GTypeInstance*)item)->g_class);
		return klass->event(item, event);
	}
	else if (pActiveTool->OnEvent(event)) return true;
	m_pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(widget), "data");
	m_pWidget = widget;
	double x, y;
	x = event->button.x;
	y = event->button.y;
	gnome_canvas_item_w2i(GNOME_CANVAS_ITEM(m_pData->Group), &x, &y);
	if (event->type == GDK_BUTTON_PRESS)
	{
		if (item == m_pData->Background)
		{
			item = NULL;
			std::map<Object*, GnomeCanvasGroup*>::iterator i = m_pData->Items.begin();
			gcpBond* pBond;
			while (i != m_pData->Items.end())
			{
				if ((*i).first->GetType() == BondType)
				{
					pBond = (gcpBond*)(*i).first;
					if (pBond->GetDist(x / m_pData->ZoomFactor, y / m_pData->ZoomFactor) < (m_DefaultPadding + m_DefaultBondWidth / 2) / m_pData->ZoomFactor)
					{
						item = GNOME_CANVAS_ITEM((*i).second);
						break;
					}
				}
				i++;
			}
		}
	}
	Object* Obj;
	switch (event->type)
	{
	case GDK_BUTTON_PRESS:
		Obj = (item) ? (Object*) g_object_get_data(G_OBJECT(item), "object") : NULL;
		switch (event->button.button)
		{
			case 1:
				if (bDragging) break;
				bDragging = true;
				pActiveTool->OnClicked(this, Obj, x, y, event->button.state);
				return true;
			case 3:
				if (Obj) Obj->ShowContextualMenu(event->button.button, event->button.time);
		}
		break;
	case GDK_MOTION_NOTIFY:
		if (!bDragging) break;
		pActiveTool->OnDrag(x, y, event->button.state);
		return true;
	case GDK_BUTTON_RELEASE:
		switch (event->button.button)
		{
		case 1:
			if (!bDragging) break;
			bDragging = false;
			pActiveTool->OnRelease(x, y, event->button.state);
			clear_status();
			return true;
		}
		break;
	}
	return false;
}

void gcpView::AddObject(Object* pObject)
{
	std::list<GtkWidget*>::iterator i;
	for (i = m_Widgets.begin(); i != m_Widgets.end(); i++)
		pObject->Add(*i);
}

void gcpView::AddBond(gcpBond* pBond)
{
	std::list<GtkWidget*>::iterator i;
	for (i = m_Widgets.begin(); i != m_Widgets.end(); i++)
		AddBond(pBond, *i);
	gcpAtom* pAtom = (gcpAtom*)pBond->GetAtom(0);
	pAtom->Update();
	Update(pAtom);
	pAtom = (gcpAtom*)pBond->GetAtom(1);
	pAtom->Update();
	Update(pAtom);
}

void gcpView::AddBond(gcpBond* pBond, GtkWidget* widget)
{
	if (!widget) return;
	gcpAtom *pAtom0, *pAtom1;
	if (!(pAtom0 = (gcpAtom*)pBond->GetAtom(0))) return; 
	if (!(pAtom1 = (gcpAtom*)pBond->GetAtom(1))) return;
	unsigned char order = pBond->GetOrder();
	if (order == 0) return;
	double x1, y1, x2, y2, x, y, dx, dy, dx1, dy1;
	int i, n;
	GnomeCanvasPoints* points = NULL;
	GnomeCanvasGroup* group;
	GnomeCanvasItem* item;
	m_pWidget = widget;
	m_pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(m_pWidget), "data");
	group = GNOME_CANVAS_GROUP(gnome_canvas_item_new(m_pData->Group, gnome_canvas_group_get_type(), NULL));
	g_signal_connect(G_OBJECT(group), "event", G_CALLBACK(on_event), m_pWidget);
	g_object_set_data(G_OBJECT(group), "object", pBond);
	switch (pBond->GetType())
	{
		case NormalBondType:
			points = gnome_canvas_points_new (2);
			i = 1;
			while (pBond->GetLine2DCoords(i++, &x1, &y1, &x2, &y2))
			{
				points->coords[0] = x1 * m_pData->ZoomFactor;
				points->coords[1] = y1 * m_pData->ZoomFactor;
				points->coords[2] = x2 * m_pData->ZoomFactor;
				points->coords[3] = y2 * m_pData->ZoomFactor;
				item = gnome_canvas_item_new(
										group,
										gnome_canvas_line_get_type(),
										"points", points,
										"fill_color", (m_pData->IsSelected(pBond))? SelectColor: Color,
										"width_units", m_pData->BondWidth,
										NULL);
				g_object_set_data(G_OBJECT(item), "object", pBond);
				g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), m_pWidget);
			}
			break;
		case UpBondType:
			points = gnome_canvas_points_new (3);
			pBond->GetLine2DCoords(1, &x1, &y1, &x2, &y2);
			points->coords[0] = x1 * m_pData->ZoomFactor;
			points->coords[1] = y1 * m_pData->ZoomFactor;
			x = sqrt(square(x2 - x1) + square(y2 - y1));
			dx = (y1 - y2) / x * m_pData->StereoBondWidth / 2;
			dy = (x2 - x1) / x * m_pData->StereoBondWidth / 2;
			points->coords[2] = x2 * m_pData->ZoomFactor + dx;
			points->coords[3] = y2 * m_pData->ZoomFactor + dy;
			points->coords[4] = x2 * m_pData->ZoomFactor - dx;
			points->coords[5] = y2 * m_pData->ZoomFactor - dy;
			item = gnome_canvas_item_new(
										group,
										gnome_canvas_polygon_get_type(),
										"points", points,
										"fill_color", (m_pData->IsSelected(pBond))? SelectColor: Color,
										NULL);
			g_object_set_data(G_OBJECT(item), "object", pBond);
			g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), m_pWidget);
			break;
		case DownBondType:
			points = gnome_canvas_points_new (4);
			pBond->GetLine2DCoords(1, &x1, &y1, &x2, &y2);
			x1 *= m_pData->ZoomFactor;
			y1 *= m_pData->ZoomFactor;
			x2 *= m_pData->ZoomFactor;
			y2 *= m_pData->ZoomFactor;
			x = sqrt(square(x2 - x1) + square(y2 - y1));
			n = int(floor(x / (m_pData->HashDist + m_pData->HashWidth)));
			dx1 = (x2 - x1) / x * m_pData->HashWidth;
			dy1 = (y2 - y1) / x * m_pData->HashWidth;
			dx = (y1 - y2) / x * m_pData->StereoBondWidth / 2;
			dy = (x2 - x1) / x * m_pData->StereoBondWidth / 2;
			points->coords[0] = x1 + dx;
			points->coords[1] = y1 + dy;
			points->coords[2] = x1 - dx;
			points->coords[3] = y1 - dy;
			dx *= (1 - m_pData->HashWidth / x);
			dy *= (1 - m_pData->HashWidth / x);
			points->coords[4] = x1 + dx1 - dx;
			points->coords[5] = y1 + dy1 - dy;
			points->coords[6] = x1 + dx1 + dx;
			points->coords[7] = y1 + dy1 + dy;
			dx = (x2 - x1) / x * (m_pData->HashDist + m_pData->HashWidth)
				- (y1 - y2) / x * m_pData->StereoBondWidth / 2 * (m_pData->HashDist + m_pData->HashWidth) / x;
			dy = (y2 - y1) / x * (m_pData->HashDist + m_pData->HashWidth)
				- (x2 - x1) / x * m_pData->StereoBondWidth / 2 *  (m_pData->HashDist + m_pData->HashWidth) / x;
			dx1 = (x2 - x1) / x * (m_pData->HashDist + m_pData->HashWidth)
				+ (y1 - y2) / x * m_pData->StereoBondWidth / 2 *  (m_pData->HashDist + m_pData->HashWidth) / x;
			dy1 = (y2 - y1) / x * (m_pData->HashDist + m_pData->HashWidth)
				+ (x2 - x1) / x * m_pData->StereoBondWidth / 2 *  (m_pData->HashDist + m_pData->HashWidth) / x;
			item = gnome_canvas_item_new(
								group,
								gnome_canvas_polygon_get_type(),
								"points", points,
								"fill_color", (m_pData->IsSelected(pBond))? SelectColor: Color,
								NULL);
			g_object_set_data(G_OBJECT(item), "object", pBond);
			g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), m_pWidget);
			for (int i = 1; i < n; i++)
			{
				points->coords[0] += dx;
				points->coords[1] += dy;
				points->coords[2] += dx1;
				points->coords[3] += dy1;
				points->coords[6] += dx;
				points->coords[7] += dy;
				points->coords[4] += dx1;
				points->coords[5] += dy1;
				item = gnome_canvas_item_new(
								group,
								gnome_canvas_polygon_get_type(),
								"points", points,
								"fill_color", (m_pData->IsSelected(pBond))? SelectColor: Color,
								NULL);
				g_object_set_data(G_OBJECT(item), "object", pBond);
				g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), m_pWidget);
			}
			break;
		case UndeterminedBondType:
			pBond->GetLine2DCoords(1, &x1, &y1, &x2, &y2);
			x1 *= m_pData->ZoomFactor;
			y1 *= m_pData->ZoomFactor;
			x2 *= m_pData->ZoomFactor;
			y2 *= m_pData->ZoomFactor;
			GnomeCanvasPathDef *path_def;
			path_def = gnome_canvas_path_def_new();
			gnome_canvas_path_def_moveto(path_def, x1, y1);
			double length;
			length = sqrt(square(x2 - x1) + square(y2 - y1));
			n = (int)length / 3;
			int	s = 1;
			dx = (x2 - x1) / n;
			dy = (y2 - y1) / n;
			x = x1;
			y = y1;
			for (int i = 1; i < n; i++)
			{
				x1 = x + dx / 3 + dy /1.5 * s;
				y1 = y + dy / 3 - dx /1.5 * s;
				dx1 = x + dx / 1.5 + dy /1.5 * s;
				dy1 = y + dy / 1.5 - dx /1.5 * s;
				x += dx;
				y += dy;
				s *= -1;
				gnome_canvas_path_def_curveto(path_def, x1, y1, dx1, dy1, x, y);
			}
			x1 = x + dx / 3 + dy /1.5 * s;
			y1 = y + dy / 3 - dx /1.5 * s;
			dx1 = x + dx / 1.5 + dy /1.5 * s;
			dy1 = y + dy / 1.5 - dx /1.5 * s;
			gnome_canvas_path_def_curveto(path_def, x1, y1, dx1, dy1, x2, y2);
			item = gnome_canvas_item_new(
										group,
										gnome_canvas_bpath_get_type(),
										"outline_color", (m_pData->IsSelected(pBond))? SelectColor: Color,
										"width_units", m_pData->BondWidth,
										"bpath", path_def,
										NULL);
			gnome_canvas_path_def_unref(path_def);
			g_object_set_data(G_OBJECT(item), "object", pBond);
			g_signal_connect(G_OBJECT(item), "event", G_CALLBACK(on_event), m_pWidget);
			break;
	}
	if (points) gnome_canvas_points_free(points);
	m_pData->Items[pBond] = group;
	if (pAtom0->GetZ() != 6)
		gnome_canvas_item_raise_to_top(GNOME_CANVAS_ITEM(m_pData->Items[pAtom0]));
	if (pAtom1->GetZ() != 6)
		gnome_canvas_item_raise_to_top(GNOME_CANVAS_ITEM(m_pData->Items[pAtom1]));
	gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(group), &x1, &y1, &x2, &y2);
	gnome_canvas_request_redraw((GnomeCanvas*)m_pWidget, (int)x1, (int)y1, (int)x2, (int)y2);
}

GtkWidget* gcpView::CreateNewWidget()
{
	gtk_widget_push_colormap(gdk_rgb_get_colormap());
	m_pWidget = gnome_canvas_gcp_new();
	gtk_widget_pop_colormap();
	GtkWidget* pWidget = (m_Widgets.size() > 0) ? m_Widgets.front() : NULL;
	if (m_pWidget)
	{
		g_object_set_data(G_OBJECT(m_pWidget), "view", this);
		g_object_set_data(G_OBJECT(m_pWidget), "doc", m_pDoc);
		m_pData = new gcpWidgetData();
		m_pData->Canvas = m_pWidget;
		g_object_set_data(G_OBJECT(m_pWidget), "data", m_pData);
		m_pData->View = this;
		gnome_canvas_set_pixels_per_unit(GNOME_CANVAS(m_pWidget), 1);
		gnome_canvas_set_scroll_region(GNOME_CANVAS(m_pWidget), 0, 0, m_width, m_height);
//		gtk_widget_set_size_request(m_pWidget, m_width, m_height);//FIXME: probably necessary when embedded
		m_pData->ZoomFactor = DefaultZoomFactor;
		m_pData->Padding = m_DefaultPadding * DefaultZoomFactor;
		m_pData->BondWidth = m_DefaultBondWidth * DefaultZoomFactor;
		m_pData->BondLength = m_DefaultBondLength * DefaultZoomFactor;
		m_pData->BondDist = m_DefaultBondDist * DefaultZoomFactor;
		m_pData->ArrowLength = m_DefaultArrowLength * DefaultZoomFactor;
		m_pData->ArrowHeadA = m_DefaultArrowHeadA * DefaultZoomFactor;
		m_pData->ArrowHeadB = m_DefaultArrowHeadB * DefaultZoomFactor;
		m_pData->ArrowHeadC = m_DefaultArrowHeadC * DefaultZoomFactor;
		m_pData->ArrowDist = m_DefaultArrowDist * DefaultZoomFactor;
		m_pData->HashWidth = m_DefaultHashWidth * DefaultZoomFactor;
		m_pData->HashDist = m_DefaultHashDist * DefaultZoomFactor;
		m_pData->StereoBondWidth = m_DefaultStereoBondWidth * DefaultZoomFactor;
		m_pData->Group = gnome_canvas_root(GNOME_CANVAS(m_pWidget));
		m_pData->Background = gnome_canvas_item_new(
									m_pData->Group,
									gnome_canvas_rect_get_type(),
									"x1", 0.0,
									"y1", 0.0,
									"x2", (double) m_width,
									"y2", (double) m_height,
									"fill_color", "white",
									NULL);
		g_signal_connect(G_OBJECT(m_pData->Background), "event", G_CALLBACK(on_event), m_pWidget);
		g_signal_connect(G_OBJECT(m_pWidget), "destroy", G_CALLBACK(on_destroy), this);
		g_signal_connect(G_OBJECT(m_pWidget), "size_allocate", G_CALLBACK(on_size), this);
		gtk_widget_show(m_pWidget);
		m_Widgets.push_back(m_pWidget);
		if (pWidget)
		{
			gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(pWidget), "data");
			std::map<Object*, GnomeCanvasGroup*>::iterator i;
			for (i = pData->Items.begin(); i != pData->Items.end(); i++)
				if ((*i).first->GetType() == BondType) AddBond((gcpBond*)((*i).first), m_pWidget);
				else (*i).first->Add(m_pWidget);
		}
		else
		{
			m_PangoContext = gtk_widget_create_pango_context(m_pWidget);
			g_object_ref(G_OBJECT(m_PangoContext));
			UpdateFont();
		}
	}
	return m_pWidget;
}

void gcpView::OnDestroy(GtkWidget* widget)
{
	if (IsEmbedded())
	{
		delete (gcpWidgetData*)g_object_get_data(G_OBJECT(widget), "data");
		m_Widgets.remove(widget);
	}
	else delete m_pDoc;
}

GnomeCanvasItem* gcpView::GetCanvasItem(GtkWidget* widget, Object* Object)
{
	gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(widget), "data");
	if ((!pData) || (pData->View != this)) return NULL;
	return (GnomeCanvasItem*) pData->Items[Object];
}

void gcpView::Print(GnomePrintContext *pc, gdouble width, gdouble height)
{
	gnome_print_gsave(pc);
	gnome_print_setrgbcolor (pc, 0, 0, 0);
	double matrix [6] = {.15, 0, 0, -.15,	//FIXME: find a reasonable scale
					0, height};
	if (!IsEmbedded())	//FIXME: add a print setup dialog and use real margins values
	{
		matrix[4] += 30;
		matrix[5] -= 30;
	}
	gnome_print_concat (pc, matrix);
	gnome_print_setlinewidth(pc, m_DefaultBondWidth);

	GList* ItemList = m_pData->Group->item_list;
	Object* pObject;
	gcpAtom* pAtom;
	gcpBond* pBond;
	GnomeFont *font, *smallfont;
	GnomeGlyphList *gl;
	gchar* buf;
	ArtDRect bbox;
	int i, n;
	double x, X[8], x1, y1, x2, y2, dx, dy, dx1, dy1;
	while (ItemList)
	{
		pObject = (Object*) g_object_get_data(G_OBJECT(ItemList->data), "object");
		if (pObject) switch (pObject->GetType())
		{
			case BondType:
				i = 1;
				pBond = (gcpBond*) pObject;
				switch(pBond->GetType())
				{
						case NormalBondType:
							while (pBond->GetLine2DCoords(i++, &x1, &y1, &x2, &y2))
							{
								gnome_print_moveto (pc, x1, y1);
								gnome_print_lineto (pc, x2, y2);
								gnome_print_stroke (pc);
							}
							break;
					case UpBondType:
						pBond->GetLine2DCoords(1, &x1, &y1, &x2, &y2);
						gnome_print_moveto(pc, x1, y1);
						x = sqrt(square(x2 - x1) + square(y2 - y1));
						dx = (y1 - y2) / x * m_DefaultStereoBondWidth / 2;
						dy = (x2 - x1) / x * m_DefaultStereoBondWidth / 2;
						gnome_print_lineto(pc, x2 + dx, y2 + dy);
						gnome_print_lineto(pc, x2 - dx, y2 - dy);
						gnome_print_closepath(pc);
						gnome_print_fill(pc);
						gnome_print_stroke (pc);
						break;
					case DownBondType:
						pBond->GetLine2DCoords(1, &x1, &y1, &x2, &y2);
						x = sqrt(square(x2 - x1) + square(y2 - y1));
						int n = int(floor(x / (m_DefaultHashDist + m_DefaultHashWidth)));
						dx1 = (x2 - x1) / x * m_DefaultHashWidth;
						dy1 = (y2 - y1) / x * m_DefaultHashWidth;
						dx = (y1 - y2) / x * m_DefaultStereoBondWidth / 2;
						dy = (x2 - x1) / x * m_DefaultStereoBondWidth / 2;
						gnome_print_moveto(pc, X[0] = x1 + dx, X[1] = y1 + dy);
						gnome_print_lineto(pc, X[2] = x1 - dx, X[3] = y1 - dy);
						dx *= (1 - m_DefaultHashWidth / x);
						dy *= (1 - m_DefaultHashWidth / x);
						gnome_print_lineto(pc, X[4] = x1 + dx1 - dx, X[5] = y1 + dy1 - dy);
						gnome_print_lineto(pc, X[6] = x1 + dx1 + dx, X[7] = y1 + dy1 + dy);
						gnome_print_closepath(pc);
						gnome_print_fill(pc);
						gnome_print_stroke (pc);
						dx = (x2 - x1) / x * (m_DefaultHashDist + m_DefaultHashWidth)
							- (y1 - y2) / x * m_DefaultStereoBondWidth / 2 * (m_DefaultHashDist + m_DefaultHashWidth) / x;
						dy = (y2 - y1) / x * (m_DefaultHashDist + m_DefaultHashWidth)
							- (x2 - x1) / x * m_DefaultStereoBondWidth / 2 *  (m_DefaultHashDist + m_DefaultHashWidth) / x;
						dx1 = (x2 - x1) / x * (m_DefaultHashDist + m_DefaultHashWidth)
							+ (y1 - y2) / x * m_DefaultStereoBondWidth / 2 *  (m_DefaultHashDist + m_DefaultHashWidth) / x;
						dy1 = (y2 - y1) / x * (m_DefaultHashDist + m_DefaultHashWidth)
							+ (x2 - x1) / x * m_DefaultStereoBondWidth / 2 *  (m_DefaultHashDist + m_DefaultHashWidth) / x;
						for (int i = 1; i < n; i++)
						{
							gnome_print_moveto(pc, X[0] += dx, X[1] += dy);
							gnome_print_lineto(pc, X[2] += dx1, X[3] += dy1);
							gnome_print_lineto(pc, X[4] += dx1, X[5] += dy1);
							gnome_print_lineto(pc, X[6] += dx, X[7] += dy);
							gnome_print_closepath(pc);
							gnome_print_fill(pc);
							gnome_print_stroke (pc);
						}
						break;
				}
				break;
			case AtomType:
				pAtom = (gcpAtom*) pObject;
				if ((pAtom->GetZ() == 6) && pAtom->GetBondsNumber()) break;
				font = gnome_font_find_closest_from_full_name((const guchar*)"Sans 48");//FIXME: calculate the real size
				if (!font) break;
				pAtom->GetCoords(&x1, &y1);
				gl = gnome_glyphlist_from_text_dumb (font, 0x000000ff, 0.0, 0.0, (const guchar*)pAtom->GetSymbol());
				gnome_glyphlist_bbox (gl, NULL, 0, &bbox);
				x2 = (bbox.x1 - bbox.x0) / 2 + m_DefaultPadding;
				y2 = (bbox.y1 - bbox.y0) / 2 + m_DefaultPadding;
				gnome_print_setrgbcolor (pc, 1.0, 1.0, 1.0);
				gnome_print_rect_filled (pc, x1 - x2, y1 - y2, 2 * x2, 2 * y2);
				gnome_print_setrgbcolor (pc, 0.0, 0.0, 0.0);
				gnome_print_moveto (pc, x1 - (bbox.x1 - bbox.x0) / 2, y1 + (bbox.y1 - bbox.y0) / 2);
				gnome_print_gsave(pc);
				gnome_print_scale(pc, 1, -1);
				gnome_print_glyphlist (pc, gl);
				gnome_print_grestore(pc);
				gnome_glyphlist_unref (gl);
				n = pAtom->GetAttachedHydrogens();
				if (n > 0)
				{
					if (pAtom->GetBestSide())
					{
						x1 += (bbox.x1 - bbox.x0) / 2 + m_DefaultPadding;//FIXME: m_DefaultPadding is not appropriate
						gl = gnome_glyphlist_from_text_dumb (font, 0x000000ff, 0.0, 0.0, (const guchar*)"H");
						gnome_print_moveto (pc, x1, y1 + (bbox.y1 - bbox.y0) / 2);
						gnome_glyphlist_bbox (gl, NULL, 0, &bbox);
						x2 = (bbox.x1 - bbox.x0) + m_DefaultPadding;
						y2 = (bbox.y1 - bbox.y0) / 2 + m_DefaultPadding;
						gnome_print_setrgbcolor (pc, 1.0, 1.0, 1.0);
						gnome_print_gsave(pc);
						gnome_print_rect_filled (pc, x1, y1 - y2, x2, 2 * y2);
						gnome_print_grestore(pc);
						gnome_print_setrgbcolor (pc, 0.0, 0.0, 0.0);
						gnome_print_gsave(pc);
						gnome_print_scale(pc, 1, -1);
						gnome_print_glyphlist (pc, gl);
						gnome_print_grestore(pc);
						gnome_glyphlist_unref (gl);
						if (n > 1)
						{
							x1 += (bbox.x1 - bbox.x0) + m_DefaultPadding;//FIXME: m_DefaultPadding is not appropriate
							buf = g_strdup_printf("%d", n);
							smallfont = gnome_font_find_closest_from_full_name((const guchar*)"Sans 32");//FIXME: calculate the real size
							gl = gnome_glyphlist_from_text_dumb (smallfont, 0x000000ff, 0.0, 0.0, (const guchar*)buf);
							g_free(buf);
							gnome_print_moveto (pc, x1, y1 + (bbox.y1 - bbox.y0) / 2 + 9.6);//FIXME: 9.6 should be replaced by a calculated value
							gnome_glyphlist_bbox (gl, NULL, 0, &bbox);
							x2 = (bbox.x1 - bbox.x0) + m_DefaultPadding;
							y2 = (bbox.y1 - bbox.y0) / 2 + m_DefaultPadding;
							gnome_print_setrgbcolor (pc, 1.0, 1.0, 1.0);
							gnome_print_gsave(pc);
							gnome_print_rect_filled (pc, x1, y1 - y2 + 9.6, x2, 2 * y2);
							gnome_print_grestore(pc);
							gnome_print_setrgbcolor (pc, 0.0, 0.0, 0.0);
							gnome_print_gsave(pc);
							gnome_print_scale(pc, 1, -1);
							gnome_print_glyphlist (pc, gl);
							gnome_print_grestore(pc);
							gnome_glyphlist_unref (gl);
						}
					}
					else
					{
						x1 -= (bbox.x1 - bbox.x0) / 2 + m_DefaultPadding;//FIXME: m_DefaultPadding is not appropriate
						if (n > 1)
						{
							buf = g_strdup_printf("%d", n);
							smallfont = gnome_font_find_closest_from_full_name((const guchar*)"Sans 32");//FIXME: calculate the real size
							gl = gnome_glyphlist_from_text_dumb (smallfont, 0x000000ff, 0.0, 0.0, (const guchar*)buf);
							g_free(buf);
							dy =  (bbox.y1 - bbox.y0) / 2 + 9.6;//FIXME: 9.6 should be replaced by a calculated value
							gnome_glyphlist_bbox (gl, NULL, 0, &bbox);
							x2 = (bbox.x1 - bbox.x0) + m_DefaultPadding;
							y2 = (bbox.y1 - bbox.y0) / 2 + m_DefaultPadding;
							x1 -= (bbox.x1 - bbox.x0);//FIXME: m_DefaultPadding is not appropriate
							gnome_print_setrgbcolor (pc, 1.0, 1.0, 1.0);
							gnome_print_rect_filled (pc, x1 - m_DefaultPadding, y1 - 2 * y2 + dy, x2, 2 * y2);
							gnome_print_setrgbcolor (pc, 0.0, 0.0, 0.0);
							gnome_print_moveto (pc, x1, y1 + dy);
							gnome_print_gsave(pc);
							gnome_print_scale(pc, 1, -1);
							gnome_print_glyphlist (pc, gl);
							gnome_print_grestore(pc);
							gnome_glyphlist_unref (gl);
							x1 -= DefaultPadding;//FIXME: m_DefaultPadding is not appropriate
						}
						gl = gnome_glyphlist_from_text_dumb (font, 0x000000ff, 0.0, 0.0, (const guchar*)"H");
						gnome_glyphlist_bbox (gl, NULL, 0, &bbox);
						gnome_glyphlist_bbox (gl, NULL, 0, &bbox);
						x2 = (bbox.x1 - bbox.x0) + m_DefaultPadding;
						y2 = (bbox.y1 - bbox.y0) / 2 + m_DefaultPadding;
						x1 -= (bbox.x1 - bbox.x0);//FIXME: m_DefaultPadding is not appropriate
						gnome_print_setrgbcolor (pc, 1.0, 1.0, 1.0);
						gnome_print_gsave(pc);
						gnome_print_rect_filled (pc, x1 - m_DefaultPadding, y1 - y2, x2, 2 * y2);
						gnome_print_grestore(pc);
						gnome_print_setrgbcolor (pc, 0.0, 0.0, 0.0);
						gnome_print_moveto (pc, x1, y1 + (bbox.y1 - bbox.y0) / 2);
						gnome_print_gsave(pc);
						gnome_print_scale(pc, 1, -1);
						gnome_print_glyphlist (pc, gl);
						gnome_print_grestore(pc);
						gnome_glyphlist_unref (gl);
					}
				}
				break;
			default:
				pObject->Print(pc);
		}
		ItemList = ItemList->next;
	}
	gnome_print_grestore(pc);
}

void gcpView::Update(Object* pObject)
{
	std::list<GtkWidget*>::iterator i;
	switch (pObject->GetType())
	{
		case AtomType:
			for (i = m_Widgets.begin(); i != m_Widgets.end(); i++)
			{
				gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(*i), "data");
				gtk_object_destroy(GTK_OBJECT(pData->Items[pObject]));
				pObject->Add(*i);//FIXME: replace with Update
			}
			break;
		case BondType:
			for (i = m_Widgets.begin(); i != m_Widgets.end(); i++)
			{
				gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(*i), "data");
				gtk_object_destroy(GTK_OBJECT(pData->Items[pObject]));
				AddBond((gcpBond*)pObject, *i);
			}
			break;
		default:
			if (pObject->HasChildren())
			{
				std::map<std::string, Object*>::iterator i;
				Object* pObj;
				for (pObj = pObject->GetFirstChild(i); pObj; pObj = pObject->GetNextChild(i)) Update(pObj);
			}
			else for (i = m_Widgets.begin(); i != m_Widgets.end(); i++)
			{
				pObject->Update(*i);
			}
			break;

	}
}

GnomeCanvasItem* gcpView::GetBackground()
{
	return m_pData->Background;
}

double gcpView::GetZoomFactor()
{
	return m_pData->ZoomFactor;
}

void gcpView::UpdateFont()
{
	pango_context_set_font_description(m_PangoContext, m_PangoFontDesc);
	PangoLayout* pl = pango_layout_new(m_PangoContext);
	PangoRectangle rect;
	pango_layout_set_text(pl, "lj", 2);
	pango_layout_get_extents(pl, &rect, NULL);
	m_dFontHeight = rect.height / PANGO_SCALE;
}
	
void gcpView::Remove(Object* pObject)
{
	std::list<GtkWidget*>::iterator i;
	for (i = m_Widgets.begin(); i != m_Widgets.end(); i++)
	{
		gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(*i), "data");
		Object* pObj = pObject->GetMolecule();
		if (pObj) pData->SelectedObjects.remove(pObj);
		else pData->SelectedObjects.remove(pObject);
		if (pData->Items[pObject]) gtk_object_destroy(GTK_OBJECT(pData->Items[pObject]));
		pData->Items.erase(pObject);
	}
}

void gcpView::UpdateLabel(const gchar* label)
{
	GtkLabel* pLabel = (GtkLabel*)g_object_get_data(G_OBJECT(m_pWidget), "label");
	if (pLabel) gtk_label_set_text(pLabel, label);
	GtkContainer* pMenu = (GtkContainer*)g_object_get_data(G_OBJECT(m_pWidget), "menu");
	if (!pMenu) return;
	GList* l = gtk_container_get_children(pMenu);
	if (GTK_IS_LABEL(l->data)) gtk_label_set_text((GtkLabel*)(l->data), label);
}

void gcpView::OnDeleteSelection(GtkWidget* w)
{
	m_pWidget = w;
	m_pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(w), "data");
	gcpWidgetData* pData;
	std::list<GtkWidget*>::iterator j;
	for (j = m_Widgets.begin(); j != m_Widgets.end(); j++)
	{
		if (*j == m_pWidget) continue;
		pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(*j), "data");
		pData->UnselectAll();
	}
	while (!m_pData->SelectedObjects.empty()) m_pDoc->Remove(m_pData->SelectedObjects.front());
	m_pDoc->FinishOperation();
	m_pData->ClearSelection();
}

void gcpView::OnCopySelection(GtkWidget* w)
{
	GtkWidget* box = gtk_message_dialog_new(NULL, (GtkDialogFlags) 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, "Sorry, not implemented in this version");
	g_signal_connect_swapped (G_OBJECT (box), "response", G_CALLBACK (gtk_widget_destroy), G_OBJECT (box));
	gtk_widget_show_all(box);
}

void gcpView::OnPasteSelection(GtkWidget* w)
{
	GtkWidget* box = gtk_message_dialog_new(NULL, (GtkDialogFlags) 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, "Sorry, not implemented in this version");
	g_signal_connect_swapped (G_OBJECT (box), "response", G_CALLBACK (gtk_widget_destroy), G_OBJECT (box));
	gtk_widget_show_all(box);
}

void gcpView::OnCutSelection(GtkWidget* w)
{
	GtkWidget* box = gtk_message_dialog_new(NULL, (GtkDialogFlags) 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, "Sorry, not implemented in this version");
	g_signal_connect_swapped (G_OBJECT (box), "response", G_CALLBACK (gtk_widget_destroy), G_OBJECT (box));
	gtk_widget_show_all(box);
}

bool gcpView::OnKeyPress(GtkWidget* w, GdkEventKey* event)
{
	if (pActiveTool->OnEvent((GdkEvent*)event)) return true;
	switch(event->keyval)
	{
		case GDK_Delete:
		case GDK_Clear:
		case GDK_BackSpace:
			OnDeleteSelection(w);
			return true;
		case GDK_Shift_L:
		case GDK_Shift_R:
			if (pActiveTool) pActiveTool->OnKeyPressed(GDK_SHIFT_MASK);
			return true;
		case GDK_Control_L:
		case GDK_Control_R:
			if (pActiveTool) pActiveTool->OnKeyPressed(GDK_CONTROL_MASK);
			return true;
		case GDK_Alt_L:
		case GDK_Alt_R:
			if (pActiveTool) pActiveTool->OnKeyPressed(GDK_MOD1_MASK);//FIXME: might be not portable
			return true;
		default:
			break;
	}
	return false;
}

bool gcpView::OnKeyRelease(GtkWidget* w, GdkEventKey* event)
{
	switch(event->keyval)
	{
		case GDK_Shift_L:
		case GDK_Shift_R:
			if (pActiveTool) pActiveTool->OnKeyReleased(GDK_SHIFT_MASK);
			return true;
		case GDK_Control_L:
		case GDK_Control_R:
			if (pActiveTool) pActiveTool->OnKeyReleased(GDK_CONTROL_MASK);
			return true;
		case GDK_Alt_L:
		case GDK_Alt_R:
			if (pActiveTool) pActiveTool->OnKeyReleased(GDK_MOD1_MASK);//FIXME: might be not portable
			return true;
		default:
			break;
	}
	return false;
}
	
bool gcpView::OnSize(int width, int height)
{
	if ((width == m_width) && (height == m_height)) return true;
	std::list<GtkWidget*>::iterator i;
	int x, y;
	m_height = height;
	m_width = width;
	for (i = m_Widgets.begin(); i != m_Widgets.end(); i++)
	{
		gnome_canvas_set_scroll_region(GNOME_CANVAS(*i), 0, 0, width, height);
		gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(*i), "data");
		if (pData->Background) g_object_set(G_OBJECT(pData->Background), "x2", (double) width, "y2", (double) height, NULL);
	}
	return true;
}

void gcpView::UpdateSize(double x1, double y1, double x2, double y2)
{
	if (x1 < 0.0) x2 -= x1;
	if (y1 < 0.0) y2 -= y1;
	list<GtkWidget*>::iterator i;
	if ((x2 > m_width) || (y2 > m_height))
		for (i = m_Widgets.begin(); i != m_Widgets.end(); i++)
			gtk_widget_set_size_request(*i,(int)ceil(x2), (int)ceil(y2));
	if ((x1 < 0.0) || (y1 < 0.0))
	{
		x1 = - x1;
		y1 = - y1;
		gcpWidgetData* pData = (gcpWidgetData*)g_object_get_data(G_OBJECT(m_pWidget), "data");
		m_pDoc->Move(x1 / pData->ZoomFactor, y1 / pData->ZoomFactor);
		Update(m_pDoc);
	}
}

void gcpView::SetGnomeCanvasRichTextActive(GnomeCanvasRichText* item)
{
	m_ActiveRichText = item;
	bDragging = false;
}

void gcpView::PrepareUnselect()
{
	if (pActiveTool) pActiveTool->NotifyViewChange();
}
